import os
from utils_ak.granular_storage.df.file.stream import StreamFile
import pandas as pd
from utils_ak.os import rename_path, remove_path


class ParquetFile(StreamFile):
    def __init__(self, fn, buffer_size=None, buffer_timeout=60, compression='snappy', engine='auto'):
        self.fn = fn
        self.swap1_fn = fn + '.swap1'
        self.swap2_fn = fn + '.swap2'
        self.engine = engine
        self.compression = compression
        super().__init__(buffer_size, buffer_timeout)

    def append(self, df, append=True):
        if append and os.path.exists(self.fn):
            old_df = self.read()
            df = pd.concat([old_df, df], axis=0)
        if not os.path.exists(self.fn):
            df.reset_index().to_parquet(self.fn, engine=self.engine, compression=self.compression)
        else:
            df.reset_index().to_parquet(self.swap1_fn, engine=self.engine, compression=self.compression)
            rename_path(self.fn, self.swap2_fn)
            rename_path(self.swap1_fn, self.fn)
            remove_path(self.swap2_fn)

    def append_stream(self, df):
        super().append_stream(df, 'default')

    def read(self):
        return pd.read_parquet(self.fn, engine=self.engine).set_index('index')



import os
import paramiko
import scp
import sys
import glob

from utils_ak.str import cast_unicode


def progress(filename, size, sent):
    sys.stdout.write("%s\'s progress: %.2f%%   \r" % (filename, float(sent) / float(size) * 100))


class SSH(paramiko.SSHClient):
    """
    Wrapper for paramiko ssh client and ssh client.
    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
        self.load_system_host_keys()
        self.set_missing_host_key_policy(paramiko.AutoAddPolicy)
        self.connect()

    def connect(self):
        super().connect(*self.args, **self.kwargs)

    def put(self, *args, **kwargs):
        with scp.SCPClient(self.get_transport(), progress=progress) as scp_cli:
            scp_cli.put(*args, **kwargs)

    def _unfold_files(self, fns):
        if isinstance(fns, str):
            res = glob.glob(fns, recursive=True)
            return [fn for fn in res if os.path.isfile(fn)]
        elif isinstance(fns, list):
            return sum([self._unfold_files(fn) for fn in fns], [])
        else:
            raise Exception('Unknown wrong file format')

    def put_missing(self, fns, remote_path):
        existing_files = self.ls(remote_path)
        existing_files = [os.path.basename(fn) for fn in existing_files]
        fns = self._unfold_files(fns)
        fns = [fn for fn in fns if os.path.basename(fn) not in existing_files]
        self.put(fns, remote_path=remote_path, recursive=True)

    def get(self, *args, **kwargs):
        with scp.SCPClient(self.get_transport(), progress=progress) as scp_cli:
            scp_cli.get(*args, **kwargs)

    def _format_command(self, command):
        if isinstance(command, list):
            return '; '.join([self._format_command(x) for x in command])
        elif isinstance(command, str):
            return command
        else:
            raise ValueError('Unknown command type')

    def exec(self, command, from_path=None, bufsize=-1, timeout=None, get_pty=False, environment=None, raw_output=False,
             raise_on_error=True):
        if from_path:
            command = [f'cd {from_path}', command]

        command = self._format_command(command)
        stdin, stdout, stderr = self.exec_command(command, bufsize, timeout, get_pty, environment)

        if raw_output:
            return stdin, stdout, stderr

        stdout, stderr = stdout.read(), stderr.read()
        stdout, stderr = cast_unicode(stdout), cast_unicode(stderr)
        if raise_on_error and stderr:
            raise Exception(f'Command failed: {stderr}')
        output = stdout + stderr
        return self._remove_last_new_line(output)

    def run(self, *args, **kwargs):
        return self.exec(*args, **kwargs)

    def _remove_last_new_line(self, s):
        if not s:
            return s
        return s[:-1] if s[-1] == '\n' else s

    def ls(self, path='.', full_path=True):
        stdin, stdout, stderr = self.exec('ls', from_path=path, raw_output=True)
        stdout = cast_unicode(stdout.read())
        stdout = self._remove_last_new_line(stdout)
        res = stdout.split('\n')
        if full_path and path != '.':
            res = [os.path.join(path, x) for x in res]
        return res


if __name__ == '__main__':
    with SSH(hostname='52.197.120.112', port=50121, username='ubuntu', key_filename=r'C:\Users\xiaomi\YandexDisk\IT\Python\jupyter2\paramiko\ssh_keys/qtb_tokyo') as ssh:
        print(ssh.exec(['cd git', 'ls']))  # == ssh.exec('cd git; ls')
        local_fn = r"C:\Users\xiaomi\YandexDisk\IT\Python\jupyter2\paramiko\test.txt"
        ssh.put(local_fn, 'remote.txt')
        print(ssh.exec('ls'))
        ssh.get('remote.txt', 'local.txt')
        print(ssh.ls('git'))

        ssh.put('local_folder', 'remote_folder', recursive=True)
        print(ssh.ls())

        import traceback

        try:
            print(ssh.exec('lsasdfasdf'))
        except:
            traceback.print_exc()

        ssh.put_missing(r"C:\Users\xiaomi\YandexDisk\IT\Python\jupyter2\paramiko\*", 'remote_folder/')




from datetime import datetime
from utils_ak.time.dt import cast_timedelta
import pandas as pd

class StreamFile(object):
    """ A simple file with buffer support of dataframes """
    def __init__(self, buffer_size=None, buffer_timeout=60):
        self.buffer_size = buffer_size or 0
        self.cur_size = 0
        self.buffer_timeout = cast_timedelta(buffer_timeout)

        # {key: [value1, value2, ...]}
        self.buffer = {}
        self.last_flush = datetime.now()

    def append(self, *args, **kwargs):
        raise NotImplementedError('Not implemented')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.flush_buffer()

    def append_stream(self, df, table):
        if len(df) == 0:
            raise Exception('Cannot add an empty dataframe')

        self.buffer.setdefault(table, []).append(df)
        self.cur_size += len(df)

        if self.is_buffer_full():
            self.flush_buffer()

    def is_buffer_full(self):
        if self.buffer_timeout:
            if datetime.now() - self.last_flush > self.buffer_timeout:
                return True
        return self.cur_size > self.buffer_size

    def flush_buffer(self):
        for table, dfs in self.buffer.items():
            df = pd.concat(dfs, axis=0)
            self.append(df, table)

        self.buffer = {}
        self.last_flush = datetime.now()
        self.cur_size = 0



from copy import deepcopy


def shatter_range(function, split, params, *args, **kwargs):
    from_name, to_name = split

    if len(params) == 2:
        from_, to_ = params
        spacing = 1
    else:
        from_, to_, spacing = params

    results = []

    for left in range(from_, to_, spacing):
        right = min(to_, left + spacing)

        kwargs.update({from_name: left, to_name: right})

        new_function = partial(function, *args, **kwargs)
        new_function.__setattr__(from_name, left)
        new_function.__setattr__(to_name, right)

        results.append(new_function)

    return results


def partial(function, *args, **kwargs):
    def partially_applied_function(*args2, **kwargs2):
        kwargs2.update(kwargs)
        return function(*(args + args2), **kwargs2)

    return partially_applied_function


def repeat(f, obj, n):
    res = obj
    for i in range(n):
        res = f(res)
    return res


if __name__ == '__main__':
    def go(left=1, right=100):
        return left, right


    ranges = shatter_range(go, split=["left", "right"], params=(1, 100, 5))
    assert ranges[1]() == (6, 11)

    go_from_5 = partial(go, left=5)
    assert go_from_5(right=10) == (5, 10)

    go_from_5 = partial(go, 5)
    assert go_from_5(10) == (5, 10)
#



import time
import os
import asyncio
from loguru import logger
from utils_ak.mongo_job_queue.worker.microservice import WorkerMicroservice
from utils_ak.serialization import cast_dict_or_list


class Worker:
    def __init__(self, id, payload, message_broker):
        self.id = id
        self.payload = payload
        self.microservice = WorkerMicroservice(id, message_broker=message_broker)

    async def process(self):
        raise NotImplemented

    def run(self):
        async def send_initial():
            await asyncio.sleep(0.1)
            await self.process()
        self.microservice.tasks.append(asyncio.ensure_future(send_initial()))
        self.microservice.run()


def run_worker(worker_cls, config=None):
    config = config or os.environ.get('CONFIG')
    assert config is not None, 'Config not specified'
    config = cast_dict_or_list(config)
    worker = worker_cls(config['worker_id'], config['payload'], config['message_broker'])
    worker.run()



from utils_ak.dag import *

from utils_ak.clock import *
from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.actors.pipe import *
from utils_ak.fluid_flow.actors.container import Container
from utils_ak.fluid_flow.calculations import *

from functools import wraps


def switch(f):
    @wraps(f)
    def inner(self, *args, **kwargs):
        pipe_switch(self, self.io_containers['in'], 'in')
        pipe_switch(self, self.io_containers['out'], 'out')

        res = f(self, *args, **kwargs)

        pipe_switch(self, self.io_containers['in'], 'in')
        pipe_switch(self, self.io_containers['out'], 'out')
        return res
    return inner


class Processor(Actor, PipeMixin):
    def __init__(self, name, items=None, processing_time=0, transformation_factor=1, max_pressures=None, limits=None):
        super().__init__(name)

        max_pressures = max_pressures or [None, None]
        limits = limits or [None, None]
        items = items or ['default', 'default']
        self.io_containers = {'in': Container('In', item=items[0], max_pressures=(max_pressures[0], None), limits=(limits[0], None)),
                           'out': Container('Out', item=items[1], max_pressures=(None, max_pressures[1]), limits=(None, limits[1]))}
        self._pipe = pipe_connect(self.io_containers['in'], self.io_containers['out'], '_pipe')
        self.processing_time = processing_time

        self.last_pipe_speed = None
        self.transformation_factor = transformation_factor

    def is_limit_reached(self, orient):
        return self.io_containers[orient].is_limit_reached(orient)

    def inner_actors(self):
        return [self.io_containers['in'], self._pipe, self.io_containers['out']]

    def on_set_pressure(self, topic, ts, event):
        self._pipe.set_pressure('out', event['pressure'], event['item'])

    def subscribe(self):
        self.event_manager.subscribe(f'processor.set_pressure.{self.id}', self.on_set_pressure)

    @switch
    def update_value(self, ts):
        self.io_containers['in'].update_value(ts)
        self.io_containers['out'].update_value(ts, factor=self.transformation_factor)

    @switch
    def update_pressure(self, ts):
        self.io_containers['in'].update_pressure(ts, orients=['in'])
        self.io_containers['out'].update_pressure(ts, orients=['out'])

    @switch
    def update_speed(self, ts):
        self.io_containers['in'].update_speed(ts, set_out_pressure=False)

        if self.processing_time == 0:
            # set new inner pressure at once
            self._pipe.set_pressure('out', self.io_containers['in'].speed('in'), self.io_containers['out'].item)
        else:
            # set inner pressure delayed with processing time
            if self.last_pipe_speed != self.io_containers['in'].speed('in'):
                self.add_event(f'processor.set_pressure.{self.id}', ts + self.processing_time, {'pressure': self.io_containers['in'].speed('in'), 'item': self.io_containers['out'].item})
                self.last_pipe_speed = self.io_containers['in'].speed('in')

        self._pipe.update_speed(ts)

        self.io_containers['out'].update_speed(ts)

    @switch
    def update_triggers(self, ts):
        for node in [self.io_containers['in'], self.io_containers['out']]:
            node.update_triggers(ts)

    def __str__(self):
        return f'Processor: {self.name}'

    def stats(self):
        return {node.name: node.stats() for node in [self.io_containers['in'], self._pipe, self.io_containers['out']]}

    def display_stats(self):
        return [self.io_containers['in'].display_stats(), self.io_containers['out'].display_stats()]

    def active_periods(self, orient='in'):
        return self.io_containers[orient].active_periods()


import logging
import logging.handlers

from utils_ak.log.extra_fields_logger import ExtraFieldsLogger

import os
import sys


def get_log_file_name():
    exec_file_path = os.path.abspath(sys.argv[0])

    exec_file_name = os.path.splitext(os.path.basename(exec_file_path))[0]
    exec_file_dir = os.path.dirname(exec_file_path)
    log_dir = os.path.join(exec_file_dir, 'logs')
    try:
        os.stat(log_dir)
    except:
        os.mkdir(log_dir)

    log_file_name = os.path.join(log_dir, f'{exec_file_name}.log')

    return log_file_name


def reset_logging():
    root_logger = logging.getLogger()
    root_logger.handlers = []


def configure_stream_logging(fmt='%(asctime)s %(name)s: %(message)s',
                             stream=True,
                             stream_level=logging.DEBUG,
                             file_level=None,
                             formatter=None,
                             file_stream=False,
                             logs_path=None,
                             stdout=False):

    file_level = file_level or stream_level

    logging.setLoggerClass(ExtraFieldsLogger)

    formatter = formatter or logging.Formatter(fmt=fmt)
    root_logger = logging.getLogger()
    root_logger.handlers = []
    root_logger.setLevel(0)
    if stream:
        handler = logging.StreamHandler(sys.stdout if stdout else None)
        handler.setLevel(stream_level)
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    if file_stream:
        logs_path = logs_path or get_log_file_name()
        fileHandler = logging.handlers.TimedRotatingFileHandler(logs_path, when='midnight', backupCount=21)
        fileHandler.setLevel(file_level)
        fileHandler.setFormatter(formatter)
        root_logger.addHandler(fileHandler)


configure_logging = configure_stream_logging

if __name__ == '__main__':
    configure_logging(file_stream=True, file_level='DEBUG', stream_level='INFO')
    logging.info('foo')
    logger = logging.getLogger('My _logging')
    logger.error('bar')
    # reset_logging()
    logger.info('foo')
    logger.error('bar')
    logger.error('bar', extra={'a': 123})
    logger.debug('debug message')


def float_pattern():
    return r'([+-]?[0-9]*[.]?[0-9]+)'


def spaces_on_edge(where='beg'):
    if where == 'beg':
        return r'^ *'
    elif where == 'end':
        return r' *$'
    else:
        raise Exception('Unknown where')


import logging
import json
from utils_ak.serialization import JsonSerializer


class SimpleCustomFieldFormatter(logging.Formatter):
    def __init__(self, fmt=None, msg_fmt="{msg} {custom}", encoder=None):
        """
        :param fmt:
        :param msg_fmt: string pattern for "msg" and "extra" fields.
        """
        self.msg_fmt = msg_fmt
        self.encoder = encoder or JsonSerializer()
        super().__init__(fmt)

    def format(self, record):
        custom = getattr(record, 'custom', {})
        record.msg = self.msg_fmt.format(**{'msg': record.msg, 'custom': self.encoder.encode(custom)})
        return super().format(record)



""" Math functionality. """
from decimal import ROUND_HALF_DOWN, ROUND_DOWN, ROUND_CEILING, ROUND_05UP, ROUND_FLOOR, ROUND_HALF_EVEN, ROUND_HALF_UP, ROUND_UP, Decimal as D

ROUND_DIC = {'nearest_half_even': ROUND_HALF_EVEN,
             'floor': ROUND_FLOOR,
             'ceil': ROUND_CEILING,
             'down': ROUND_DOWN,
             'up': ROUND_UP,
             'nearest_half_up': ROUND_HALF_UP,
             'nearest_half_down': ROUND_HALF_DOWN,
             'down_except_0_and_5': ROUND_05UP}


def decimal_round(a, b, rounding='nearest_half_even', precision=0, strip=False):
    a, b = D(a), D(b)
    n = a / b / D('.1') ** precision
    rounding = ROUND_DIC.get(rounding, rounding)
    n = n.quantize(D('1.'), rounding=rounding)
    res = n * b * D('0.1') ** precision
    if strip:
        res = strip_decimal(res)
    return res


# be careful to use this for precise values: rounding 5. to 1. with floor method may result in 4 due to python floating system (5. == 4.99999999999) in python and stuff
def custom_round(a, b, rounding='nearest_half_even', precision=0):
    n = a / b / 0.1 ** precision
    n = int(decimal_round(n, 1, rounding))
    return n * b * 0.1 ** precision


def cast_decimal(f, precision=15, strip=True):
    d = decimal_round(f, D('1.0'), precision=precision)
    if strip:
        s = str(d)
        s = strip_zeros(s)
        d = D(s)
    return d


def strip_decimal(d):
    return D(strip_zeros(str(d)))


def strip_zeros(s):
    return s.rstrip('0').rstrip('.') if '.' in s else s


if __name__ == '__main__':
    roundings = [ROUND_HALF_DOWN, ROUND_DOWN, ROUND_CEILING, ROUND_05UP, ROUND_FLOOR, ROUND_HALF_EVEN, ROUND_HALF_UP, ROUND_UP]
    roundings += list(ROUND_DIC.keys())

    for rounding in roundings:
        for v in [1.99, 1.05, -1.99, -1.05]:
            print(rounding, v, 0.1, custom_round(v, 0.1, rounding))

    # 1.99
    print(decimal_round('1.99', '0.01', rounding='ROUND_FLOOR'))
    # may be 1.98 since 1.99 may be 1.9899999999
    print(decimal_round(1.99, '0.01', rounding='ROUND_FLOOR', precision=0))
    print(decimal_round(1.99, '0.01', rounding='ROUND_FLOOR', precision=2))
    # certainly 1.99, since we round 1.9899999 to 1.99 first
    print(decimal_round(1.99, '0.0100', rounding='ROUND_FLOOR', precision=10))
    print(decimal_round(1.99, '0.0100', rounding='ROUND_FLOOR', precision=10))
    print(decimal_round(1.99, '0.0100', rounding='ROUND_FLOOR', precision=10, strip=True))



import anyconfig
import copy
import os
import tempfile
from loguru import logger
from utils_ak.deployment.config import BASE_DIR
from utils_ak.deployment.controller import Controller
from utils_ak.serialization import cast_js, cast_dict_or_list
from utils_ak.os import *
from utils_ak.dict import fill_template

# todo: success and failure responses (or errors?)

class DockerController(Controller):
    def start(self, deployment):
        id = deployment['id']

        assert len(deployment['containers']) == 1, "Only one-container pods are supported for now"

        entity, container = list(deployment['containers'].items())[0]
        params = {'entity': entity, 'deployment_id': id, 'image': container['image']}

        config = cast_dict_or_list(os.path.join(BASE_DIR, 'example/docker-compose.yml.template'))
        config = fill_template(config, **params)

        for k, v in container['command_line_arguments'].items():
            config['services'][entity]['command'].append(f'--{k}')
            config['services'][entity]['command'].append(cast_js(v))

        makedirs(f'data/docker-compose/{id}/')
        fn = f'data/docker-compose/{id}/docker-compose.yml'

        with open(fn, 'w') as f:
            f.write(anyconfig.dumps(config, 'yaml'))

        execute(f'docker-compose -f "{fn}" up -d --no-build')

    def stop(self, deployment_id):
        for id in self._get_docker_ids(deployment_id):
            execute(f'docker stop {id}')
            execute(f'docker rm {id}')

        remove_path(f'data/docker-compose/{deployment_id}/')

    def _get_docker_ids(self, deployment_id):
        ids = execute(f'docker ps -q -f name={deployment_id}*').split('\n')
        ids = [id for id in ids if id]
        return ids

    def log(self, deployment_id):
        ids = self._get_docker_ids(deployment_id)
        for id in ids:
            logger.debug('Logs for id', id=id, logs=execute(f'docker logs {id}'))


def test_docker_controller():
    import anyconfig
    import time
    from utils_ak.loguru import logger, configure_loguru_stdout
    configure_loguru_stdout('DEBUG')

    deployment = anyconfig.load('../../example/deployment.yml')
    ctrl = DockerController()
    ctrl.start(deployment)
    time.sleep(3)
    ctrl.log(deployment['id'])
    ctrl.stop(deployment['id'])


if __name__ == '__main__':
    test_docker_controller()


import asyncio
from utils_ak.mongo_job_queue.worker.worker import Worker
from utils_ak.mongo_job_queue.worker.worker_test import *


class TestWorker(Worker):
    async def process(self):
        if self.payload.get('type') == 'batch':
            await asyncio.sleep(1)
            self.microservice.send_state('running', {})
            for i in range(5):
                self.microservice.send_state('running', {'progress': (i + 1) * 20})
                await asyncio.sleep(0.1)
            self.microservice.send_state('success', {'response': '42'})
            self.microservice.stop()
        elif self.payload.get('type') == 'streaming':
            await asyncio.sleep(3)
            self.microservice.send_state('running', {})

            while True:
                self.microservice.send_state('running', {'foo': 'bar'})
                await asyncio.sleep(3)
        else:
            raise Exception(f'Bad payload type {self.payload.get("type")}')


def test_batch():
    test_worker(TestWorker, {'type': 'batch'}, message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}}))


def test_streaming():
    test_worker(TestWorker, {'type': 'streaming'}, message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}}))


def test_deployment():
    test_worker_deployment('deployment.yml', message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}}))


if __name__ == '__main__':
    # test_batch()
    # test_streaming()
    test_deployment()


from utils_ak.granular_storage.df.gran_df.sync import GranularSync
from utils_ak.time.dt import *


class GranularStreamer(object):
    def __init__(self, gs, path, buffer_timeout=24 * 3600, level=3, stream_level=4, init_path=False, migration_expected=True):
        self.gs = gs
        self.path = path
        self.stream_path = path + '_stream'
        self.migration_expected = migration_expected

        self.gran = gs.get_granular(path, level=level)

        # todo: make level properly
        self.level = level
        self.stream_level = stream_level
        self.init_stream_gran(mode='a')

        self.last_sync = datetime.now()
        self.buffer_timeout = buffer_timeout

        if init_path and not self.gs.exists(path):
            self.gs.get_granular(path, level=self.level)

    def init_stream_gran(self, mode):
        self.gran_stream = self.gs.get_granular(self.stream_path, level=self.stream_level, mode=mode)

    def flush(self):
        self.gran_stream.flush()

    def append_stream(self, df, key):
        self.gran_stream.append_stream(df, key)
        if cast_sec((datetime.now() - self.last_sync)) > self.buffer_timeout:
            self.last_sync = datetime.now()
            self.sync_with_base()

    def sync_with_base(self):
        print('Sync with base')
        # base granular not exists yet. Waiting for migration
        if not self.gs.exists(self.path):
            print('path not exists')
            return

        gran_base = self.gs[self.path]
        # waiting for migration to be complete before syncing
        if self.migration_expected and not gran_base.metadata.get('migration_complete'):
            print('Migration not complete yet')
            return

        print('Writing!')
        # handle opened files and current buffer
        self.gran_stream.flush()
        self.gran_stream.close()

        sync = GranularSync(self.gs, self.gs)
        sync.sync(self.stream_path, self.path)

        # initialize new streamer with write mode - to remove current stream granular_storage
        self.init_stream_gran(mode='w')


def ex_stream():
    print('-' * 100)
    print('Stream test')
    print('-' * 100)
    import time
    from utils_ak.granular_storage.df.gran_df import GranularMsgPackStorage as Storage
    from utils_ak.os_tools import remove
    remove('granular_storage/gs_root/')
    gs = Storage('granular_storage/gs_root')

    index = [cast_datetime('2018.05.01') + timedelta(hours=6) * i for i in range(3)]
    df = pd.DataFrame(np.random.randn(3, 1), index=index, columns=['0'])

    streamer = GranularStreamer(gs, 'sample_path', buffer_timeout=3, init_path=True, migration_expected=False)
    streamer.append_stream(df.iloc[:1], 'sample_table')

    def print_table():
        try:
            print(gs['sample_path']['sample_table'])
        except:
            print('No table yet')

    time.sleep(4)
    streamer.append_stream(df.iloc[1:2], 'sample_table')
    print_table()
    time.sleep(1)
    streamer.append_stream(df.iloc[2:3], 'sample_table')
    print_table()
    streamer.flush()
    print_table()
    streamer.sync_with_base()
    print_table()


def ex_migration():
    print('-' * 100)
    print('Migration test')
    print('-' * 100)
    from utils_ak.granular_storage.df.gran_df.sync import GranularSync
    from utils_ak.granular_storage.df.gran_df import GranularMsgPackStorage as Storage
    from utils_ak.os_tools import remove
    remove('granular_storage/gs_from/')
    remove('granular_storage/gs_to/')
    gs_from = Storage('granular_storage/gs_from')
    gs_to = Storage('granular_storage/gs_to')

    index = [cast_datetime('2018.05.01') + timedelta(hours=6) * i for i in range(3)]
    df = pd.DataFrame(np.random.randn(3, 1), index=index, columns=['0'])

    streamer = GranularStreamer(gs_to, 'sample_path', buffer_timeout=0, init_path=True, migration_expected=True)

    def print_table():
        try:
            print(gs_to['sample_path']['sample_table'])
        except:
            print('No table yet')

    streamer.append_stream(df.iloc[1:2], 'sample_table')
    print('Before migration')
    print_table()

    gran_from = gs_from.get_granular('sample_path', level=0)
    gran_from.append(df.iloc[:2], 'sample_table')

    # make migration happen
    sync = GranularSync(gs_from, gs_to)
    sync.sync('sample_path', 'sample_path', flag_migration=True)
    print('After migration')
    print_table()
    streamer.append_stream(df.iloc[2:3], 'sample_table')
    print('Stream streamed after migration')
    print_table()


if __name__ == '__main__':
    ex_stream()
    ex_migration()



from .broker import Broker
from utils_ak.message_queue.clients.zmq_client import ZMQClient

ENDPOINT_TYPES = ['sub', 'pub']


class ZMQBroker(Broker):
    def __init__(self, endpoints):
        """
        :param endpoints: {collection_name: type, ...}
        """

        self.endpoints = endpoints

        # {endpoint: collection}
        self.collections = {}

        for collection, config in endpoints.items():
            if config['type'] in ['pub', 'sub']:
                self.collections[config['endpoint']] = collection
            else:
                # forwarder
                self.collections[config['endpoint_pub']] = collection
                self.collections[config['endpoint_sub']] = collection

        self.cli = ZMQClient()

        self.async_supported = False

    def _get_endpoint_info(self, collection, action):
        """
        :param collection: str
        :param action: sub or pub
        :return:
        """
        config = self.endpoints[collection]
        endpoint_type = config['type']
        if endpoint_type in ['pub', 'sub']:
            endpoint = config['endpoint']
            conn_type = 'bind' if endpoint_type == action else 'connect'
        elif endpoint_type == 'forwarder':
            if action == 'pub':
                endpoint = config['endpoint_pub']
            else:
                endpoint = config['endpoint_sub']
            conn_type = 'connect'
        else:
            raise Exception(f'Unsupported endpoint type {endpoint_type}')
        return endpoint, conn_type

    def subscribe(self, collection, topic):
        endpoint, conn_type = self._get_endpoint_info(collection, 'sub')
        self.cli.subscribe(endpoint, topic, conn_type=conn_type)

    def publish(self, collection, topic, msg):
        endpoint, conn_type = self._get_endpoint_info(collection, 'pub')
        self.cli.publish(endpoint, topic, msg, conn_type=conn_type)

    def poll(self, timeout=0.):
        msg = self.cli.poll(timeout)
        if not msg:
            return
        endpoint, topic, msg = msg
        return self.collections[endpoint], topic, msg



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['utils_ak',
 'utils_ak.architecture',
 'utils_ak.architecture.finite_state_machine',
 'utils_ak.architecture.func',
 'utils_ak.architecture.integrity_guard',
 'utils_ak.architecture.operation',
 'utils_ak.architecture.sink',
 'utils_ak.async_tools',
 'utils_ak.aws',
 'utils_ak.block_tree',
 'utils_ak.block_tree.pushers',
 'utils_ak.builtin',
 'utils_ak.callback_timer',
 'utils_ak.clock',
 'utils_ak.color',
 'utils_ak.config',
 'utils_ak.dag',
 'utils_ak.deployment',
 'utils_ak.deployment.controller',
 'utils_ak.deployment.controller.controllers',
 'utils_ak.deployment.example',
 'utils_ak.dict',
 'utils_ak.exceptions',
 'utils_ak.fluid_flow',
 'utils_ak.fluid_flow.actors',
 'utils_ak.fluid_flow.tests',
 'utils_ak.functions',
 'utils_ak.git',
 'utils_ak.golang',
 'utils_ak.granular_storage',
 'utils_ak.granular_storage.df',
 'utils_ak.granular_storage.df.file',
 'utils_ak.granular_storage.df.gran_df',
 'utils_ak.granular_storage.storage',
 'utils_ak.granular_storage.table',
 'utils_ak.granular_storage.table.timeseries',
 'utils_ak.iteration',
 'utils_ak.json',
 'utils_ak.jupyter',
 'utils_ak.kafka',
 'utils_ak.log',
 'utils_ak.loguru',
 'utils_ak.message_queue',
 'utils_ak.message_queue.brokers',
 'utils_ak.message_queue.brokers.examples',
 'utils_ak.message_queue.clients',
 'utils_ak.messenger',
 'utils_ak.messenger.slack',
 'utils_ak.mongo_job_queue',
 'utils_ak.mongo_job_queue.worker',
 'utils_ak.mongo_job_queue.worker.test',
 'utils_ak.mongoengine',
 'utils_ak.naming',
 'utils_ak.networking',
 'utils_ak.networking.proxy',
 'utils_ak.nn',
 'utils_ak.numeric',
 'utils_ak.openpyxl',
 'utils_ak.os',
 'utils_ak.pandas',
 'utils_ak.pandas_writer',
 'utils_ak.plot',
 'utils_ak.portion',
 'utils_ak.properties',
 'utils_ak.properties.archive',
 'utils_ak.proxy',
 'utils_ak.re',
 'utils_ak.reflexion',
 'utils_ak.script_manager',
 'utils_ak.scripts',
 'utils_ak.serialization',
 'utils_ak.simple_event_manager',
 'utils_ak.simple_microservice',
 'utils_ak.simple_microservice.examples',
 'utils_ak.simple_vector',
 'utils_ak.split_file',
 'utils_ak.ssh',
 'utils_ak.state',
 'utils_ak.state.provider',
 'utils_ak.str',
 'utils_ak.streaming',
 'utils_ak.streaming.windowing',
 'utils_ak.system',
 'utils_ak.time',
 'utils_ak.tqdm',
 'utils_ak.zmq',
 'utils_ak.zmq.benchmark']

package_data = \
{'': ['*'],
 'utils_ak.deployment': ['controller/controllers/data/123/*'],
 'utils_ak.log': ['logs/*'],
 'utils_ak.mongo_job_queue': ['data/docker-compose/601fb3b2fae079694e9bffc6/*',
                              'data/docker-compose/601fb5e86d7bcff9e4bb2e8e/*',
                              'data/docker-compose/601fb62c3b2e4db2e22bcb5d/*',
                              'data/docker-compose/601fb75839b1782336a7526a/*',
                              'data/docker-compose/601fb83c47748754f032ed2f/*',
                              'data/docker-compose/601fb8b4ecb47ccfedca6ff8/*',
                              'data/docker-compose/601fb90f5a68d5fb8bcb1917/*',
                              'data/docker-compose/601fb945e561bd6721cf1c4f/*',
                              'data/docker-compose/601fba9228c92e888a8e0688/*',
                              'data/docker-compose/601ffb46bb447437ca90d7bf/*',
                              'data/docker-compose/6020028c0098c1f51119cc55/*',
                              'data/docker-compose/602129cf1cd00b24cb27b31c/*',
                              'data/kubernetes/601fbb5002f5d3c83008580d/*',
                              'data/kubernetes/6020033d6a02e816b4b114a7/*',
                              'data/kubernetes/60212a51a66ecefe66120c0a/*',
                              'data/kubernetes/60212c15c601175c48bf8a53/*',
                              'data/kubernetes/60212d461d8bad556bd388ae/*',
                              'data/kubernetes/60212deec77a775d132a6f48/*',
                              'data/kubernetes/60212e2b078151797a094266/*',
                              'data/kubernetes/60212e6aa5b45b4bd9e16de1/*',
                              'data/kubernetes/602137b92ff441fe7bdf2bcc/*',
                              'data/kubernetes/602137ba2ff441fe7bdf2bcd/*',
                              'data/kubernetes/602137d82ff441fe7bdf2bce/*',
                              'data/kubernetes/6021393630b8552bbe7ba5b8/*',
                              'data/kubernetes/6021394530b8552bbe7ba5b9/*',
                              'data/kubernetes/602139f8ef7d97c515cdff53/*',
                              'data/kubernetes/6021648d23146081340fa8a8/*',
                              'data/kubernetes/602164c923146081340fa8a9/*',
                              'worker/test/data/docker-compose/deployment_id/*']}

install_requires = \
['Delorean>=1.0.0,<2.0.0',
 'GitPython>=3.1.9,<4.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'anyconfig>=0.9.11,<0.10.0',
 'boto3>=1.15.18,<2.0.0',
 'empyrical>=0.5.5,<0.6.0',
 'filelock>=3.0.12,<4.0.0',
 'fire>=0.4.0,<0.5.0',
 'h5py>=2.10.0,<3.0.0',
 'joblib>=0.17.0,<0.18.0',
 'loguru>=0.5.3,<0.6.0',
 'mongoengine>=0.22.1,<0.23.0',
 'msgpack>=1.0.2,<2.0.0',
 'numpy>=1.19.2,<2.0.0',
 'opencv-python>=4.4.0,<5.0.0',
 'openpyxl>=3.0.6,<4.0.0',
 'pandas>=1.1.3,<2.0.0',
 'paramiko>=2.7.2,<3.0.0',
 'pathlib>=1.0.1,<2.0.0',
 'pickleDB>=0.9.2,<0.10.0',
 'portion>=2.1.4,<3.0.0',
 'pymongo>=3.11.0,<4.0.0',
 'pyperclip>=1.8.1,<2.0.0',
 'pyzmail36>=1.0.4,<2.0.0',
 'pyzmq>=19.0.2,<20.0.0',
 'requests-futures>=1.0.0,<2.0.0',
 'retrypy>=0.0.33,<0.0.34',
 'scp>=0.13.2,<0.14.0',
 'seaborn>=0.11.0,<0.12.0',
 'slacker>=0.14.0,<0.15.0',
 'sortedcollections>=2.1.0,<3.0.0',
 'stackprinter>=0.2.5,<0.3.0',
 'telegram>=0.0.1,<0.0.2',
 'tenacity>=6.2.0,<7.0.0',
 'tqdm>=4.50.2,<5.0.0']

setup_kwargs = {
    'name': 'utils-ak',
    'version': '0.1.3',
    'description': 'Lots of small utils',
    'long_description': None,
    'author': 'akadaner',
    'author_email': 'arseniikadaner@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



""" Storing and retrieving granular_storage in the filesystem . """

from utils_ak.granular_storage.df.file import *
from utils_ak.granular_storage.df.gran_df import GranularMsgPackStorage, GranularParquetStorage

GRAN_LEVEL_ZERO = 0
GRAN_LEVEL_YEAR = 1
GRAN_LEVEL_MONTH = 2
GRAN_LEVEL_DAY = 3
GRAN_LEVEL_HOUR = 4

MainStorage = GranularMsgPackStorage
#
# if __name__ == '__main__':
#     storage = MainStorage('.')
#     gran_df = storage.get_granular('some/path/to/granular', level=GRAN_LEVEL_ZERO)
#     gran_df.append(df, append=False)



import anyconfig
import copy
import os
import tempfile

from loguru import logger

from utils_ak.deployment.config import BASE_DIR
from utils_ak.deployment.controller import Controller
from utils_ak.serialization import cast_js, cast_dict_or_list
from utils_ak.os import *
from utils_ak.dict import fill_template


class KubernetesController(Controller):
    def start(self, deployment):
        id = deployment['id']
        assert len(deployment['containers']) == 1, "Only one-container pods are supported for now"

        # create docker-compose file and run it without building
        entity, container = list(deployment['containers'].items())[0]
        params = {'entity': entity, 'deployment_id': id, 'image': container['image']}

        configs = cast_dict_or_list(os.path.join(BASE_DIR, 'example/kubernetes.deployment.yml.template'))
        configs = [fill_template(config, **params) for config in configs]

        for config in configs:
            if config['kind'] == 'ConfigMap':
                for k, v in container['command_line_arguments'].items():
                    config['data'][k.upper()] = cast_js(v)
        config_str = '---\n'.join([anyconfig.dumps(config, 'yaml') for config in configs])

        makedirs(f'data/kubernetes/{id}/')
        fn = f'data/kubernetes/{id}/kubernetes.yml'

        with open(fn, 'w') as f:
            f.write(config_str)

        execute(f'kubectl apply -f "{fn}"')

    def stop(self, deployment_id):
        fn = f'data/kubernetes/{deployment_id}/kubernetes.yml'
        execute(f'kubectl delete -f "{fn}"')
        # remove_path(f'data/kubernetes/{deployment_id}/')

    def log(self, deployment_id):
        logger.debug('Logs', logs=execute(f'kubectl logs -l deployment_id={deployment_id}'))


def test_kubernetes_controller():
    import anyconfig
    import time
    from utils_ak.loguru import logger, configure_loguru_stdout
    configure_loguru_stdout('DEBUG')

    deployment = anyconfig.load('../../example/deployment.yml')
    ctrl = KubernetesController()
    ctrl.start(deployment)
    time.sleep(10)
    ctrl.log(deployment['id'])
    ctrl.stop(deployment['id'])


if __name__ == '__main__':
    test_kubernetes_controller()


import pandas as pd
import fnmatch
import os

from utils_ak.os import execute


class ScriptManager(object):
    def __init__(self, config, prefix='_'):
        self.prefix = prefix

        self.config = {self.cast_name(name): path for name, path in config.items()}

    def start(self, name, cmd):
        execute("screen -S {} -dm bash -c \" {}\"".format(name, cmd), is_async=True)

    def ls(self, by='time'):
        output = execute('screen -ls', is_async=False)
        lines = output.split('\n')[1:-2]
        values = []
        for line in lines:
            value = line.split('\t')[1:]
            value = value[0].split('.', 1) + value[1:]
            values.append(value)
        df = pd.DataFrame(values, columns=['id', 'name', 'time', 'status'])
        df['status'] = df['status'].apply(lambda s: s.replace('(', '').replace(')', ''))
        df['time'] = df['time'].apply(lambda s: s.replace('(', '').replace(')', ''))
        df['time'] = pd.to_datetime(df['time'])
        df = df[['time', 'name', 'id', 'status']]
        df = df.sort_values(by=by)
        df = df[df['name'].str.startswith(self.prefix)]
        df = df.reset_index(drop=True)
        return df

    def kill(self, pattern):
        df = self.ls()
        if len(df) == 0:
            return
        df['full_name'] = df['id'].astype(str) + '.' + df['name']
        df = df[df['name'].apply(lambda name: fnmatch.fnmatch(name, pattern))]
        df = df[df['name'] != 'jupyter']
        for i, row in df.iterrows():
            print('Killing: {}'.format(row['full_name']))
            self.kill_one(row['full_name'])

    def restart(self, pattern):
        self.kill(pattern)
        self.run(pattern)

    def kill_one(self, name):
        if name == 'jupyter':
            raise Exception('jupyter cannot be killed for safety measures')
        return execute("screen -S " + name + " -X quit", is_async=False)

    def run_python(self, name, path):
        if not os.path.exists(path):
            raise Exception('File does not exist {}'.format(path))
        self.start(name, 'python {}'.format(path))

    def run(self, pattern=None):
        config = self.config
        df = self.ls()
        if pattern:
            config = {k: v for k, v in config.items() if fnmatch.fnmatch(k, pattern)}
        for screen_name, path in config.items():
            if screen_name in df['name'].values:
                print(f'Already running: {screen_name}')
            else:
                print(f'Running: {screen_name}')
                self.run_python(screen_name, path)

    def cast_name(self, name):
        return f'{self.prefix}{name}'

    def state(self, pattern=None):
        config = self.config
        df = self.ls()
        if pattern:
            config = {k: v for k, v in self.config.items() if fnmatch.fnmatch(k, pattern)}
        for screen_name, path in config.items():
            if screen_name not in df['name'].values:
                df.loc[len(df)] = [None, screen_name, None, 'Not running']
        return df.sort_values(by='name').reset_index(drop=True)



import io
import sys
import os
import logging
import traceback

from logging import currentframe, _srcfile


class ExtraFieldsLogger(logging.Logger):
    """ Logger with additional kwargs arguments. """

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1, **kwargs):
        """
        Low-level _logging routine which creates a LogRecord and then calls
        all the handlers of this _logging to handle the record.
        """

        sinfo = None
        if _srcfile:
            # IronPython doesn't track Python frames, so findCaller raises an
            # exception on some versions of IronPython. We trap it here so that
            # IronPython can use _logging.
            try:
                fn, lno, func, sinfo = self.findCaller(stack_info, stacklevel)
            except ValueError:  # pragma: no cover
                fn, lno, func = "(unknown file)", 0, "(unknown function)"
        else:  # pragma: no cover
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info:
            if isinstance(exc_info, BaseException):
                exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()

        if kwargs:
            extra = extra or {}
            extra.update(kwargs)

        record = self.makeRecord(self.name, level, fn, lno, msg, args, exc_info, func, extra, sinfo)
        self.handle(record)

    def findCaller(self, stack_info=False, stacklevel=1):
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = currentframe()
        # On some versions of IronPython, currentframe() returns None if
        # IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back
        orig_f = f
        while f and stacklevel > 1:
            f = f.f_back
            stacklevel -= 1
        if not f:
            f = orig_f
        rv = "(unknown file)", 0, "(unknown function)", None
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            if filename == _srcfile:
                f = f.f_back
                continue
            sinfo = None
            if stack_info:
                sio = io.StringIO()
                sio.write('Stack (most recent call last):\n')
                traceback.print_stack(f, file=sio)
                sinfo = sio.getvalue()
                if sinfo[-1] == '\n':
                    sinfo = sinfo[:-1]
                sio.close()
            rv = (co.co_filename, f.f_lineno, co.co_name, sinfo)
            break
        return rv



from utils_ak.granular_storage.storage.fs_storage import GranularTable, Storage, DB

from utils_ak.granular_storage.table.timeseries.fs_json import FsJsonTimeSeriesTable
from utils_ak.granular_storage.table.timeseries.fs_msgpack import FsMsgPackTimeSeriesTable


class JsonGranularTable(GranularTable):
    @property
    def table_cls(self):
        return FsJsonTimeSeriesTable


class MsgPackGranularTable(GranularTable):
    @property
    def table_cls(self):
        return FsMsgPackTimeSeriesTable


class JsonGranularStorage(Storage):
    @property
    def table_cls(self):
        return JsonGranularTable

    @property
    def db_cls(self):
        return DB

    @property
    def ext(self):
        return '.txt'


class MsgPackGranularStorage(Storage):
    @property
    def table_cls(self):
        return MsgPackGranularTable

    @property
    def db_cls(self):
        return DB

    @property
    def ext(self):
        return '.msg'


if __name__ == '__main__':
    from datetime import datetime, timedelta

    storage = JsonGranularStorage('/tmp/')
    db = storage.get_db('foo_db')
    table = db.get_table('foo_table', pattern='%Y%m%d_{base}')

    dt = datetime.now()
    values = [[dt.timestamp(), 'key', {'a': i}] for i in range(3)]
    dt += timedelta(days=1)
    values += [[dt.timestamp(), 'key', {'b': i}] for i in range(3)]
    dt += timedelta(days=1)
    values += [[dt.timestamp(), 'key', {'c': i}] for i in range(3)]

    for i in range(3):
        table.store(datetime.now().timestamp(), 'key', {'a': i})

    table.store_many(values, safe=False, overwrite=True)



import datetime
import msgpack
import decimal
from utils_ak.serialization.binary import pack_datetime, pack_decimal, pack_time, unpack_datetime, unpack_decimal, unpack_time, pack_date, unpack_date

EXT_DATE = 1
EXT_DATETIME = 2
EXT_DECIMAL = 3
EXT_TIME = 4


def default(obj):
    if isinstance(obj, datetime.datetime):
        return msgpack.ExtType(EXT_DATETIME, pack_datetime(obj))
    elif isinstance(obj, datetime.date):
        return msgpack.ExtType(EXT_DATE, pack_date(obj))
    elif isinstance(obj, datetime.time):
        return msgpack.ExtType(EXT_TIME, pack_time(obj))
    elif isinstance(obj, decimal.Decimal):
        return msgpack.ExtType(EXT_DECIMAL, pack_decimal(obj))
    else:
        # Wuh-woh
        raise TypeError(u'Cannot encode value of type {} to MessagePack: {}'.format(type(obj).__name__, obj))


def ext_hook(code, data):
    """
    Decodes our custom extension types
    """
    if code == EXT_DATETIME:
        return unpack_datetime(data)
    elif code == EXT_DATE:
        return unpack_date(data)
    elif code == EXT_TIME:
        return unpack_time(data)
    elif code == EXT_DECIMAL:
        return unpack_decimal(data)
    else:
        raise TypeError(u'Cannot decode unknown extension type {} from MessagePack'.format(code))


class MsgPackSerializer:
    @staticmethod
    def encode(obj):
        return msgpack.dumps(obj, default=default, use_bin_type=True)

    @staticmethod
    def decode(obj):
        return msgpack.loads(obj, ext_hook=ext_hook, raw=False)


if __name__ == '__main__':
    ser = MsgPackSerializer()
    obj = {'int': 1,
           'datetime': datetime.datetime.now(),
           'decimal': decimal.Decimal('0.12431234123000'),
           'date': datetime.date(2018, 1, 1),
           'time': datetime.datetime.now().time()}
    data = ser.encode(obj)
    print(obj)
    print(len(data), data)
    print(ser.decode(data))



"""
Formats:
    -   schemed str: scheme://ip:port or scheme://user:pass@user:pswd. Example: socks5://localhost:5556.
    -   dict: {'ip', 'port', 'scheme'} or {'ip', 'port', 'user', 'pswd', 'scheme'}. Scheme is 'http' by default.
    -   None
    Default format: dict.
"""

import re
from utils_ak import re as re_tools

# from here: https://stackoverflow.com/questions/106179/regular-expression-to-match-dns-hostname-or-ip-address
IP_PAT = "(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])"
HOST_PAT = "(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])"
PORT_PAT = '(\d+)'
SCHEME_PAT = 'http|https|socks5'
USER_PAT = '[a-zA-Z0-9]+'
PSWD_PAT = '[^\s@]+'

PROXY_PAT = f'%SCHEME://(%USER:%PSWD@)?%HOST:%PORT'
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, '%SCHEME', 'scheme', SCHEME_PAT)
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, '%USER', 'user', USER_PAT)
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, '%PSWD', 'pswd', PSWD_PAT)
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, '%HOST', 'host', HOST_PAT)
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, '%PORT', 'port', PORT_PAT)


def cast_dict(proxy_obj):
    if proxy_obj is None:
        return None
    elif isinstance(proxy_obj, dict):
        if any(x not in proxy_obj for x in ['scheme', 'host', 'port']):
            raise ValueError('Missing some proxy keys')
        if ('user' in proxy_obj or 'pswd' in proxy_obj) and any(x not in proxy_obj for x in ['user', 'pswd']):
            raise ValueError('Missing some proxy keys')
        return proxy_obj
    elif isinstance(proxy_obj, str):
        search = re.search(PROXY_PAT, proxy_obj)
        if not search:
            raise ValueError(f'Non-valid proxy {proxy_obj}')
        return search.groupdict()
    else:
        raise ValueError('Unknown proxy type')


def cast_str(proxy_obj):
    proxy = cast_dict(proxy_obj)
    if not proxy:
        raise Exception('Cannot cast proxy string from null proxy')
    elif proxy.get('user'):
        scheme, user, pswd, host, port = proxy['scheme'], proxy['user'], proxy['pswd'], proxy['host'], proxy['port']
        return f'{scheme}://{user}:{pswd}@{host}:{port}'
    else:
        scheme, host, port = proxy['scheme'], proxy['host'], proxy['port']
        return f'{scheme}://{host}:{port}'


def cast_requests_proxies(proxy_obj):
    """
    :return:
       proxies = {
        'http': 'http://10.10.1.10:3128',
        'https': 'http://10.10.1.10:1080',
    }
    """
    if not proxy_obj:
        return None
    proxy_str = cast_str(proxy_obj)
    return {'http': proxy_str, 'https': proxy_str}


if __name__ == '__main__':
    print(cast_dict('http://ip:1234'))
    print(cast_dict({'scheme': 'http', 'host': 'host', 'port': 9999}))
    print(cast_str({'scheme': 'http', 'host': 'host', 'port': 9999, 'user': 'user', 'pswd': 'pswd'}))
    print(cast_requests_proxies('http://user:pswd@host:9999'))



from utils_ak.fluid_flow import *


def test_pipe_switch_1():
    from utils_ak.fluid_flow.actors.container import Container
    ci1 = Container('I1')
    ci2 = Container('I2')
    co1 = Container('O1')
    co2 = Container('O2')

    pipe_connect(ci1, co1, '1')
    pipe_connect(ci2, co2, '2')

    for node in ci1.iterate():
        print(node)
    for node in ci2.iterate():
        print(node)
    print()

    pipe_switch(co1, co2, 'in')

    for node in ci1.iterate():
        print(node)
    for node in ci2.iterate():
        print(node)
    print()

    pipe_switch(co1, co2, 'in')
    for node in ci1.iterate():
        print(node)
    for node in ci2.iterate():
        print(node)
    print()

    pipe_switch(ci1, ci2, 'out')

    for node in ci1.iterate():
        print(node)
    for node in ci2.iterate():
        print(node)
    print()


def test_pipe_switch_2():
    from utils_ak.fluid_flow.actors.container import Container

    ci1 = Container('I1')
    co1 = Container('O1')
    co2 = Container('O2')

    pipe_connect(ci1, co1)

    for node in ci1.iterate():
        print(node)
    print()
    pipe_switch(co1, co2, 'in')

    for node in ci1.iterate():
        print(node)
    print()

if __name__ == '__main__':
    test_pipe_switch_1()
    test_pipe_switch_2()


import logging

from utils_ak.zmq import endpoint
from utils_ak.simple_microservice import SimpleMicroservice, run_listener_async
from utils_ak.log import configure_stream_logging
configure_stream_logging(stream_level=logging.INFO)


class Publisher(SimpleMicroservice):
    def __init__(self, *args, **kwargs):
        super().__init__('Publisher', *args, **kwargs)
        self.add_timer(self.timer_function, 2)

    def timer_function(self):
        self.publish_json('collection', '', {})


if __name__ == '__main__':
    run_listener_async('collection', message_broker=('zmq',{'endpoints': {'collection': {'type': 'sub', 'endpoint': endpoint('localhost', 6554)}}} ))
    Publisher( message_broker=('zmq',{'endpoints': {'collection': {'type': 'sub', 'endpoint': endpoint('localhost', 6554)}}} )).run()



import os
import pandas as pd
from datetime import timedelta

from utils_ak.granular_storage.df.file.stream import StreamFile
from utils_ak.time.dt import cast_datetime
from utils_ak.os import rename_path, remove_path

import numpy as np


# todo: safe renaming and removing?


class MsgPackFile(StreamFile):
    def __init__(self, fn, buffer_size=None, buffer_timeout=60):
        self.fn = fn
        self.swap1_fn = fn + '.swap1'
        self.swap2_fn = fn + '.swap2'
        super().__init__(buffer_size, buffer_timeout)

    def append(self, df, append=True):
        if append and os.path.exists(self.fn):
            old_df = self.read()
            df = pd.concat([old_df, df], axis=0, sort=False)
        if not os.path.exists(self.fn):
            df.reset_index().to_msgpack(self.fn)
        else:
            df.reset_index().to_msgpack(self.swap1_fn)
            rename_path(self.fn, self.swap2_fn)
            rename_path(self.swap1_fn, self.fn)
            remove_path(self.swap2_fn)

    def append_stream(self, df):
        super().append_stream(df, 'default')

    def read(self):
        return pd.read_msgpack(self.fn).set_index('index')

    def ext(self):
        return '.msg'


if __name__ == '__main__':
    index = [cast_datetime('2018.05.01') + timedelta(hours=6) * i for i in range(3)]
    df = pd.DataFrame(np.random.randn(3, 1), index=index, columns=['0'])
    file = MsgPackFile('granular_storage/test.msg')

    print(file.read())
    file.append(df)
    print(file.read())




def to_lower_name(upper_case_name):
    res = ''
    for l in upper_case_name:
        if not res:
            if l.isupper():
                res += l.lower()
            else:
                res += l
        else:
            if l.isupper():
                res += '_'
                res += l.lower()
            else:
                res += l
    return res


if __name__ == '__main__':
    print(to_lower_name('camelCase')) # camel_case
    print(to_lower_name('UpperCase')) # upper_case



from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.actors.pipe import PipeMixin, Pipe
from utils_ak.fluid_flow.calculations import *


class Hub(Actor):
    def __init__(self, name):
        super().__init__(name)

    def update_pressure(self, ts):
        # todo: hardcode. Need updated children pressures. Better solution? Update pressure top-down?
        for pipe in self.children:
            pipe.child.update_pressure(ts)

        parent_items = [pipe.current_item for pipe in self.parents]
        assert len(set(parent_items)) == 1
        item = parent_items[0]
        children_with_item = [pipe for pipe in self.children if pipe.current_item == item]

        if any(pipe.pressures['in'] is None for pipe in children_with_item):
            for pipe in self.parents:
                pipe.pressures['in'] = None
        else:
            total_output_pressure = sum(pipe.pressures['in'] for pipe in children_with_item)
            left = total_output_pressure

            for pipe in self.parents:
                pipe.pressures['in'] = nanmin([left, pipe.pressures['out']])
                left -= nanmin([left, pipe.pressures['out']])

    def update_speed(self, ts):
        assert all(isinstance(pipe, Pipe) for pipe in self.parents + self.children)

        parent_items = [pipe.current_item for pipe in self.parents]
        assert len(set(parent_items)) == 1
        item = parent_items[0]
        children_with_item = [pipe for pipe in self.children if pipe.current_item == item]

        total_input_speed = sum(pipe.current_speed for pipe in self.parents)
        left = total_input_speed

        for pipe in self.children:
            if pipe in children_with_item:
                pipe.pressures['out'] = nanmin([left, pipe.pressures['in']])
                left -= nanmin([left, pipe.pressures['in']])
            else:
                pipe.pressures['out'] = 0

    def __str__(self):
        return f'Hub: {self.name}'




import time
import logging
import multiprocessing


from utils_ak.log import configure_logging
from utils_ak.simple_microservice import SimpleMicroservice, run_listener_async
from utils_ak.mongo_job_queue.worker.test import TestWorker
from utils_ak.mongo_job_queue.monitor import MonitorActor

BROKER = 'zmq'
BROKER_CONFIG = {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}, 'monitor_out': {'endpoint': 'tcp://localhost:5556', 'type': 'sub'}}}
MESSAGE_BROKER = (BROKER, BROKER_CONFIG)


def run_monitor():
    configure_logging(stream_level=logging.DEBUG)
    ms = SimpleMicroservice('Monitor', message_broker=MESSAGE_BROKER)
    actor = MonitorActor(ms)
    ms.run()


def run_worker():
    configure_logging(stream_level=logging.DEBUG)
    worker = TestWorker('WorkerId', {'type': 'batch'}, message_broker=MESSAGE_BROKER)
    worker.run()


def test():
    configure_logging(stream_level=logging.DEBUG)
    run_listener_async('monitor_out', message_broker=MESSAGE_BROKER)
    time.sleep(1)

    multiprocessing.Process(target=run_monitor).start()
    time.sleep(3)
    multiprocessing.Process(target=run_worker).start()


if __name__ == '__main__':
    test()

